#include "cfg.h"
#include <stdio.h>
#include <assert.h>

/*
 * Lattice graph (example)
 *
 *    1 (start)
 *    |
 *    V
 *    2 ---> 3
 *    |      |
 *    V      V
 *    4 ---> 5
 *    |      |
 *    V      V
 *    6 ---> 7
 *           |
 *           V
 *           8 (end)
 *
 */

Cfg *lattice_cfg (void)
{
  Cfg *cfg;
  int i, n, N, E;

  printf ("lattice size? ");
  scanf ("%d", &n);
  assert (n > 0);

  N = 2*n+2;
  E = 3*n+1;

  cfg = init_cfg (N, E);

  /* connect nodes to form lattice */
  connect_nodes (cfg, 2, 3);
  for (i=1; i<n; i++)
    {
      connect_nodes (cfg, 2*i, 2*i+2);
      connect_nodes (cfg, 2*i+1, 2*i+3);
      connect_nodes (cfg, 2*i+2, 2*i+3);
    }

  connect_nodes (cfg, 1, 2);
  connect_nodes (cfg, N-1, N);

  connect_nodes (cfg, 1, N);

  return cfg;
}

/* repeat-until loop nest */

Cfg *repeat_cfg (void)
{
  Cfg *cfg;
  int i, n, N, E;

  printf ("repeat-loop nesting depth? ");
  scanf ("%d", &n);
  assert (n > 0);

  N = 2*n+2;
  E = 3*n+2;

  cfg = init_cfg (N, E);

  /* connect linear chain */
  for (i=1; i<N; i++)
    {
      connect_nodes (cfg, i, i+1);
    }

  /* connect backedges */
  for (i=0; i<n; i++)
    {
      connect_nodes (cfg, n+2+i, n+1-i);
    }

  connect_nodes (cfg, 1, N);

  return cfg;
}

Cfg *tarjan_cfg (void)
{
  Cfg *cfg;
  int N, E;

  N = 15;
  E = 24;
  cfg = init_cfg (N, E);

  fprintf(stderr, "test cfg from Tarjan's dominator paper\n");

  connect_nodes (cfg, 1, 2);
  connect_nodes (cfg, 2, 3);
  connect_nodes (cfg, 2, 4);
  connect_nodes (cfg, 2, 5);
  connect_nodes (cfg, 3, 6);
  connect_nodes (cfg, 4, 3);
  connect_nodes (cfg, 4, 6);
  connect_nodes (cfg, 4, 7);
  connect_nodes (cfg, 5, 8);
  connect_nodes (cfg, 5, 9);
  connect_nodes (cfg, 6, 14);
  connect_nodes (cfg, 7, 10);
  connect_nodes (cfg, 8, 11);
  connect_nodes (cfg, 9, 11);
  connect_nodes (cfg, 9, 12);
  connect_nodes (cfg, 10, 7);
  connect_nodes (cfg, 10, 13);
  connect_nodes (cfg, 11, 13);
  connect_nodes (cfg, 12, 11);
  connect_nodes (cfg, 13, 2);
  connect_nodes (cfg, 13, 11);
  connect_nodes (cfg, 13, 15);
  connect_nodes (cfg, 14, 10);

  connect_nodes (cfg, 1, N);

  return cfg;
}

Cfg *tarjan2_cfg (void)
{
  Cfg *cfg;
  int N, E;

  N = 7;
  E = 10;
  cfg = init_cfg (N, E);

  fprintf(stderr, "test cfg from Tarjan's dominator paper\n");

  connect_nodes (cfg, 1, 2);
  connect_nodes (cfg, 1, 3);
  connect_nodes (cfg, 2, 4);
  connect_nodes (cfg, 3, 2);
  connect_nodes (cfg, 3, 5);
  connect_nodes (cfg, 4, 6);
  connect_nodes (cfg, 5, 6);
  connect_nodes (cfg, 6, 5);
  connect_nodes (cfg, 6, 7);

  connect_nodes (cfg, 1, N);

  return cfg;
}

Cfg *tarjan4_cfg (void)
{
  Cfg *cfg;
  int N, E;

  N = 8;
  E = 11;
  cfg = init_cfg (N, E);

  fprintf(stderr, "test cfg from Tarjan's paper\n");

  connect_nodes (cfg, 1, 2);
  connect_nodes (cfg, 1, 3);
  connect_nodes (cfg, 2, 4);
  connect_nodes (cfg, 3, 2);
/*  connect_nodes (cfg, 3, 4); */
  connect_nodes (cfg, 3, 5);
  connect_nodes (cfg, 4, 6);
  connect_nodes (cfg, 5, 7);
  connect_nodes (cfg, 6, 7);
  connect_nodes (cfg, 7, 5);
  connect_nodes (cfg, 7, 8);

  connect_nodes (cfg, 1, N);

  return cfg;
}

Cfg *tarjan3_cfg (void)
{
  Cfg *cfg;
  int N, E;

  N = 9;
  E = 14;
  cfg = init_cfg (N, E);

  fprintf(stderr, "test cfg from Tarjan's paper\n");

  connect_nodes (cfg, 1, 2);
  connect_nodes (cfg, 2, 3);
  connect_nodes (cfg, 2, 4);
  connect_nodes (cfg, 2, 9);
  connect_nodes (cfg, 3, 5);
  connect_nodes (cfg, 4, 3);
  connect_nodes (cfg, 4, 5);
  connect_nodes (cfg, 4, 6);
  connect_nodes (cfg, 5, 7);
  connect_nodes (cfg, 6, 8);
  connect_nodes (cfg, 7, 8);
  connect_nodes (cfg, 8, 6);
  connect_nodes (cfg, 8, 9);

  connect_nodes (cfg, 1, N);

  return cfg;
}

Cfg *simple_cfg (void)
{
  Cfg *cfg;
  int N, E;

  N = 6;
  E = 9;
  cfg = init_cfg (N, E);

  fprintf(stderr, "simple test cfg\n");

  connect_nodes (cfg, 1, 2);
  connect_nodes (cfg, 2, 3);
  connect_nodes (cfg, 2, 4);
  connect_nodes (cfg, 3, 5);
  connect_nodes (cfg, 4, 5);
  connect_nodes (cfg, 5, 2);
  connect_nodes (cfg, 5, 6);

  connect_nodes (cfg, 1, N);

  return cfg;
}


Cfg *pingali_cfg(void)
{
  Cfg *cfg;
  int N, E;

  N = 8;
  E = 13;
  cfg = init_cfg (N, E);

  fprintf(stderr, "pingali's test cfg\n");

  connect_nodes (cfg, 1, 2);
  connect_nodes (cfg, 2, 3);
  connect_nodes (cfg, 2, 4);
  connect_nodes (cfg, 3, 4);
  connect_nodes (cfg, 3, 7);
  connect_nodes (cfg, 4, 5);
  connect_nodes (cfg, 4, 6);
  connect_nodes (cfg, 5, 3);
  connect_nodes (cfg, 5, 8);
  connect_nodes (cfg, 6, 8);
  connect_nodes (cfg, 7, 2);

  connect_nodes (cfg, 1, N);

  return cfg;
}

Cfg *rcj_cfg(void)
{
  Cfg *cfg;
  int N, E;

  N = 9;
  E = 12;
  cfg = init_cfg (N, E);

  fprintf(stderr, "rcj's test cfg\n");

  connect_nodes (cfg, 1, 2);
  connect_nodes (cfg, 2, 3);
  connect_nodes (cfg, 2, 4);
  connect_nodes (cfg, 3, 4);
  connect_nodes (cfg, 4, 5);
  connect_nodes (cfg, 4, 6);
  connect_nodes (cfg, 5, 7);
  connect_nodes (cfg, 6, 7);
  connect_nodes (cfg, 7, 3);
  connect_nodes (cfg, 7, 8);
  connect_nodes (cfg, 8, 9);

  connect_nodes (cfg, 1, N);

  return cfg;
}

Cfg *simple_cfg2(void)
{
  Cfg *cfg;
  int  N, E;

  N = 6;
  E = 9;
  cfg = init_cfg (N, E);

  N=6;
  assert(N<NMAX);
  printf("simple test cfg #2 (hi != ahi)\n");

  connect_nodes (cfg, 1, 2);
  connect_nodes (cfg, 2, 3);
  connect_nodes (cfg, 3, 4);
  connect_nodes (cfg, 4, 2);
  connect_nodes (cfg, 4, 5);
  connect_nodes (cfg, 5, 3);
  connect_nodes (cfg, 5, 6);

  connect_nodes (cfg, 1, N);


  return cfg;
}

Cfg *simple_cfg3(void)
{
  Cfg *cfg;
  int N, E;

  N = 4;
  E = 8;
  cfg = init_cfg (N, E);

  fprintf(stderr, "simple test cfg #3 (irreducible)\n");

  connect_nodes (cfg, 1, 2);
  connect_nodes (cfg, 1, 3);
  connect_nodes (cfg, 1, 4);
  connect_nodes (cfg, 2, 3);
  connect_nodes (cfg, 3, 4);
  connect_nodes (cfg, 3, 2);

  connect_nodes (cfg, 1, N);


  return cfg;
}

Cfg *simple_cfg4(void)
{
  Cfg *cfg;
  int N, E;

  N = 15;
  E = 27;
  cfg = init_cfg (N, E);

  fprintf(stderr, "simple test cfg #4\n");

  connect_nodes (cfg, 1, 2);
  connect_nodes (cfg, 1, 12);
  connect_nodes (cfg, 1, 14);
  connect_nodes (cfg, 2, 3);
  connect_nodes (cfg, 3, 4);
  connect_nodes (cfg, 3, 8);
  connect_nodes (cfg, 4, 5);
  connect_nodes (cfg, 4, 13);
  connect_nodes (cfg, 5, 6);
  connect_nodes (cfg, 5, 12);
  connect_nodes (cfg, 6, 7);
  connect_nodes (cfg, 6, 2);
  connect_nodes (cfg, 7, 8);
  connect_nodes (cfg, 7, 15);
  connect_nodes (cfg, 8, 3);
  connect_nodes (cfg, 8, 9);
  connect_nodes (cfg, 9, 10);
  connect_nodes (cfg, 9, 13);
  connect_nodes (cfg, 10, 8);
  connect_nodes (cfg, 10, 9);
  connect_nodes (cfg, 10, 11);
  connect_nodes (cfg, 11, 12);
  connect_nodes (cfg, 12, 15);
  connect_nodes (cfg, 13, 15);
  connect_nodes (cfg, 14, 15);

  connect_nodes (cfg, 1, N);


  return cfg;
}

/*  read in a CFG from a file
    Here is a sample CFG with 6 nodes and 7 edges
    START is always labeled 1, END is always labeled last 
    The code adds an extra edge from START to END. 

      START
       |
       v
       2
      /|\
     / | \
    v  |  v
    3  |  4
    \  |  /
     \ | /
      vvv
       5      
       |
       v
      END 

    Here is what should be there in the file. 
    First line is number of nodes, second line is number of edges
    Then, list edges by source node number, terminating each 
    list of destination nodes by 0. So each line is of the form
    sourceNode: destinationNode1 destinationNode2 .... 0
    For our example, the file would contain the following:
    -----------------------
    N: 6 
    E: 7

    1: 2 0
    2: 3 4 5 0
    3: 5 0
    4: 5 0
    5: 6 0
    -----------------------
*/


Cfg *read_cfg (void)
{
  Cfg *cfg;
  int i, N, E;

  char filename[64];
  FILE *fpi;

  fprintf(stderr, "read test cfg from file.\n\n");

  printf ("cfg filename: ");
  scanf ("%s", filename);

  fpi = fopen (filename, "r");
  if (fpi == NULL)
    {
      fprintf (stderr, "cannot open cfg file: %s\n", filename);
      return;
    }

  fscanf (fpi, "N: %d\nE: %d\n", &N, &E);
  cfg = init_cfg (N, E+1);
  
  printf("Nodes: %d\nEdges: %d\n", N, E); 

  for (i=1; i<=N; i++)
    {
      int s, d;

      fscanf (fpi, "%d:", &s);
      assert (s == i);

      fscanf (fpi, " %d", &d);
      while (d)
	{
	  connect_nodes (cfg, s, d);
	  fscanf (fpi, " %d", &d);
	}
    }

  connect_nodes (cfg, 1, N);


  return cfg;
}

