/* Control flow graph routines */

#include <assert.h>
#include "cfg.h"

Cfg *init_cfg (int _N, int _E)
{
  Cfg *cfg = (Cfg *) malloc (sizeof (Cfg));

  assert (_N  < NMAX && _E < EMAX);

  cfgEI (cfg) = 0;

  (cfg)->N = _N;
  (cfg)->E = _E;

  cfgNODES (cfg) = (Node *) malloc ((_N+1) * sizeof (Node));
  cfgEDGES (cfg) = (Edge *) malloc ((_E+1) * sizeof (Edge));

  /* initialize node, edge arrays to NULL */
  bzero (cfgNODES (cfg), (_N+1) * sizeof (Node));
  bzero (cfgEDGES (cfg), (_E+1) * sizeof (Edge));

  /* Set globals. For SPEC benchmarks, NumNodes will be changed later. */
  NumNodes = _N;
  NumEdges = _E;
  ENDNum = _N;

  return cfg;
}

void free_cfg (Cfg *cfg)
{
  free(cfgNODES (cfg));
  free(cfgEDGES(cfg));
  free(cfg);
}

void connect_nodes (Cfg * cfg, int src, int dst)
{
  int ei;
  Edge *e;

  cfgEI (cfg) += 1;
  assert (0 < cfgEI (cfg) && cfgEI (cfg) <= NumEdges);
  assert (0 < src && src <= ENDNum);
  assert (0 < dst && dst <= ENDNum);

  ei = cfgEI (cfg);
  e = &cfgEDGE(cfg, ei);

  eINDEX (e) = ei;
  eSRC (e) = src;
  eNEXTSUCC (e) = nSUCCS (cfgNODE(cfg,src));
  nSUCCS (cfgNODE(cfg,src)) = e;

  eDST (e) = dst;
  eNEXTPRED (e) = nPREDS (cfgNODE(cfg,dst));
  nPREDS (cfgNODE(cfg,dst)) = e;
}

/* end of cfg.c */
