/* conds.c */

/*
 *
 *   CONDS algorithm: phases kept separate for clarity
 *  
 */

#include <stdio.h>
#include <assert.h>
#include "cfg.h"
#include "set.h"
#include "tree.h"

static int checksum;

extern int Hi[NMAX];   /* Set here and used in cdequiv computation */
extern int bfs[NMAX];  /* node indices in BFS order */
extern int boundary[NMAX]; /*Used to define zones*/

static int lni;
extern List lnp[EMAX];  /* list node pool */
extern List *L[NMAX], *Lend[NMAX];

static int SetSize;

/****************************Preprocessing******************************/

void CondsPreprocessing (double alpha)
{
  int i;
  int a[NMAX], b[NMAX], t[NMAX], z[NMAX],nxtboundary[NMAX];

  lni = 0;  /* no list nodes used */

  for (i=1; i<=ENDNum; i++) /*EndNum is largest node number */
    {
      b[i] = t[i] = 0;
      Hi[i] = i;
      L[i] = Lend[i] = NULL;
    }

  /* for each path... */
  for (i=1; i<=npaths; i++)
    {
      b[path[i].b]++;
      t[path[i].t]++;

      /* Set Hi to highest node reached by edge out of path[i].b 
         Used in cdequiv computation. */
      if (ANCESTOR(path[i].t, Hi[path[i].b]))
	Hi[path[i].b] = path[i].t;
    };

  /* determine boundary nodes */
  for (i=NumNodes; i>=1; i--)
    {
      int u, v = bfs[i];
      a[v] = b[v] - t[v];
      z[v] = 1;
      for (u=pdom_child[v]; u; u=pdom_sibling[u])
	{
	  a[v] += a[u];
	  z[v] += z[u];
	}

      if (pdom_child[v] == 0 || (double) z[v]  > alpha*(double)a[v] + 1.0)
	{
	  boundary[v] = TRUE;
	  z[v] = 1;
	}
      else
	boundary[v] = FALSE;
    }

  /* Determine nxtboundary nodes */
  for (i=1; i<=NumNodes; i++)
    {
      int v = bfs[i];
      if (pdom[v] == 0)
	nxtboundary[v] = -1;
      else
	{
	  if (boundary[pdom[v]])
	    nxtboundary[v] = pdom[v];
	  else
	    nxtboundary[v] = nxtboundary[pdom[v]];
	}
    }

  /*Insert each path into the list at each boundary node on the path */
  for (i=1; i<=npaths; i++)
    {
      int ti = path[i].t;
      int wi = path[i].b;
      while (wi > 0 && ANCESTOR(ti,wi) && (ti != wi))
	{
	  /* add path[i] to L[wi] */
	  List *l = &lnp[lni++];  /* get list node */
	  assert (lni > 0 && lni < EMAX);
	  l->next = NULL;
	  l->pi = &path[i];
	  if (L[wi] == NULL)
	    {
	      L[wi] = Lend[wi] = l;
	    }
	  else
	    {
	      Lend[wi]->next = l;
	      Lend[wi] = l;
	    }
	  wi = nxtboundary[wi];
	}
    }
}

int CacheSize()
{return lni;
 }



/*****************************************CondsQueries***************************/

void CondsVisit (int qn, int vn, FILE *fpo, BOOLEAN printp)
{
  int c;
  List *l;

  for (l=L[vn]; l; l=l->next)
    {
      Path *pi = l->pi;
      int t = pi->t;
      
      if (ANCESTOR(t, qn) && (t != qn ))
	{
          SetSize++;
	  if (printp) fprintf (fpo, "%d->%d  ", eSRC (pi->l), eDST (pi->l));
	  checksum = (checksum + eINDEX (pi->l)) % 16000;
	}
      else
	break;
    }
  
  if (!boundary[vn])
    {
      for (c=pdom_child[vn]; c; c=pdom_sibling[c])
	{
	  CondsVisit (qn, c, fpo, printp);
	 }
    }
}

/*What is conds(n)? */
void CondsQuery (int n, FILE *fpo, BOOLEAN printp)
{
  int c;
  List *l;

  SetSize = 0;
  checksum = 0;
  if (printp) fprintf (fpo,"CONDS(%2d) = { ", n);      

  /*Inline the call to CondsVisit, as suggested in paper */
  for (l=L[n]; l; l=l->next)
    {
      Path *pi = l->pi;
      SetSize++;

      if (printp) fprintf (fpo, "%d->%d  ", eSRC (pi->l), eDST (pi->l));
      checksum = (checksum + eINDEX (pi->l)) % 16000; /* compute a checksum */
    }
  
  if (!boundary[n])
    {
      for (c=pdom_child[n]; c; c=pdom_sibling[c])
	{
	  CondsVisit (n, c, fpo, printp);
	}
    }

  if (printp) fprintf (fpo," }\n");
}




int CondsSize()
{ return SetSize;
}
