/* main.c */

#include <stdio.h>
#include <assert.h>
#include <math.h>
#include <sys/times.h>
#include <unistd.h>
#include "cfg.h"
#include "stack.h"
#include "set.h"
#include "tree.h" 

/*Number of nodes, number of edges and ID of END in CFG */
int NumNodes, NumEdges,ENDNum;

/*Storage for tree */
int pdom[NMAX], pdom1[NMAX], pdom2[NMAX];
int pdom_child[NMAX], pdom_sibling[NMAX];
static Node *node;

/*Storage for Path Array*/
Path path[NMAX];
int npaths;

/*Storage for APT and temporaries*/
int region[NMAX];
int numregions = 0;
int bfs[NMAX];  /* node indices in BFS order */
int boundary[NMAX]; /*Used to define zones*/
List lnp[EMAX];  /* list node pool */
List *L[NMAX], *Lend[NMAX];
int Hi[NMAX];  /* Set by conds preprocessing and used in cdequiv preprocessing */

/*Set bfs numbers for bottom-up computations */
static int bfs_i;
void visit_bfs(int u)
{
  int v;

  assert (u != 0);
  
  for (v=pdom_child[u]; v; v=pdom_sibling[v])
    bfs[++bfs_i] = v;

  for (v=pdom_child[u]; v; v=pdom_sibling[v])
    visit_bfs(v);
}

void ConstructRomanChariots (Cfg * cfg)
{
  int j;

  node = cfgNODES (cfg);

  /* get BFS order of nodes in tree */
  bfs_i = 0;
  bfs[++bfs_i] = ENDNum;
  visit_bfs(ENDNum);
  
  npaths = 0;
  for (j=1; j<=NumNodes; j++)
    {
      int u = bfs[j];
      Edge *ed;

      /* for each edge u->v do */
      for (ed = nSUCCS (node[u]); ed; ed = eNEXTSUCC (ed))
	{
	  int v = eDST (ed);
	  
	  if (v != pdom[u])
	    {
	      npaths++;  assert (npaths < NMAX);
	      path[npaths].b = v;
	      path[npaths].t = pdom[u];
	      path[npaths].l = ed;
	    }
	}
    }
}

int NumRoutes()
{
  return npaths;
}


void main (int argc, char *argv[])
{
  Cfg *cfg, *cfg_expanded;
  int n, i, t;

  if (argc != 2)
    {
      fprintf (stderr, "\nUsage: %s N\n\n", argv[0]);
      fprintf (stderr, "g - select graph\n");
      fprintf (stderr, "\t0 - lattice example\n");
      fprintf (stderr, "\t1 - nested repeat-until\n");
      fprintf (stderr, "\t2 - tarjan_cfg\n");
      fprintf (stderr, "\t3 - simple_cfg\n");
      fprintf (stderr, "\t4 - pingali_cfg\n");
      fprintf (stderr, "\t5 - rcj_cfg\n");
      fprintf (stderr, "\t6 - simple_cfg2\n");
      fprintf (stderr, "\t7 - simple_cfg3\n");
      fprintf (stderr, "\t8 - simple_cfg4\n");
      fprintf (stderr, "\t9 - read test cfg from file\n");
      fprintf (stderr, "\tLook at file build-cfg.c for details of these CFGs.\n");

      exit (0);
    }

  init_INT_SET ();
  init_S ();

  n = atoi (argv[1]);
/* build the CFG */
  switch (n)
    {
    case 0: cfg = lattice_cfg (); break;
    case 1: cfg = repeat_cfg ();  break;
    case 2: cfg = tarjan2_cfg ();  break;
    case 3: cfg = simple_cfg ();  break;
    case 4: cfg = pingali_cfg (); break;
    case 5: cfg = rcj_cfg (); break;
    case 6: cfg = simple_cfg2 (); break;
    case 7: cfg = simple_cfg3 (); break;
    case 8: cfg = simple_cfg4 (); break;
    case 9: cfg = read_cfg ();    break;

    default:
      fprintf (stderr, "unknown CFG kind.\n");
      exit (1);
    };

/* perform the experiment passing cfg,alpha,stderr and verbose? boolean */    

    experiment(cfg, 1.0, stderr,TRUE); /*Use 1.0 as alpha */
    free_cfg(cfg);

  exit (0);
}

/* end of main.c */
