#include <stdio.h>
#include <assert.h>
#include <memory.h>
#include "cfg.h"
#include "stack.h"
#include "tree.h"




static Stack *FloorList[NMAX]; 
static StackNode *CellDeletionList[NMAX]; /*List of entries to be deleted */


void CdequivPreprocessing( )
{
  int i;

  memset(CellDeletionList, 0, (ENDNum+1) * sizeof (StackNode *));
  

  for (i = NumNodes; i>= 1; i--)   /*perform bottom-up walk using bfs numbers*/
    { int Hi_c, u, v, Hi_1,Hi_2;
      List *l;
      Stack *FloorListTemp;

      Hi_1 = Hi_2 = 0;
      v = bfs[i]; 
      FloorListTemp = create_S();

      /* Visit children and compute FloorList,Hi_1 and Hi_2 */
      for (u = pdom_child[v]; u; u = pdom_sibling[u])
	{ 
	  if (sISEMPTY(FloorListTemp))
	      FloorListTemp = FloorList[u];
	  else 
	    combine_S (FloorList[u], FloorListTemp);

	  /* update Hi_1, Hi_2 */
	  Hi_c = Hi[u]; 
	  if (Hi_1 == 0 || ANCESTOR(Hi_c, Hi_1))
	    {
	      Hi_2 = Hi_1;
	      Hi_1 = Hi_c;
	    }
	  else if (Hi_2 == 0 || ANCESTOR(Hi_c, Hi_2))
	    {
	      Hi_2 = Hi_c;
	    };
	};

      /* Add capping route [v, Hi_2) */
      if (Hi_2 && ANCESTOR(Hi_2, Hi[v]))
	{ 
	  StackNode *sn = push_S (FloorListTemp, v);
	  snQNEXT(sn) = CellDeletionList[Hi_2];
	  CellDeletionList[Hi_2] = sn;
	};

      /* Set Hi field for node v */
      if (Hi_1 && (ANCESTOR(Hi_1, Hi[v])))
	Hi[v] = Hi_1;


      /* Push v once for each route whose bottom node is v*/
      for (l = L[v];l;l = l->next)
	{ Path *pi = l->pi;
	  int b = pi->b;

	  if (b == v) /* Check that route cached at v actually starts at v*/
	    { 
	      StackNode *sn = push_S (FloorListTemp, v);
  	      snQNEXT(sn) = CellDeletionList[pi->t];
	      CellDeletionList[pi->t] = sn;
	    }
	};

  
      /* Delete bottom nodes of routes whose top node is v*/
      while (CellDeletionList[v])
	{
	  StackNode *sn = CellDeletionList[v];
	  CellDeletionList[v] = snQNEXT(sn);
	  remove_S (sn, FloorListTemp);	
	}



      /*Store FloorList[v] and determine region for node v */
      FloorList[v] = FloorListTemp;
      if (sSIZE(FloorListTemp) == 0) 
	region[v] = 0; /* EmptyClass is 0*/
      else 
	{
	  if (snSIZE(sTOP (FloorListTemp)) != sSIZE(FloorListTemp))
	    { 
	      snSIZE (sTOP (FloorListTemp)) = sSIZE (FloorListTemp);
	      snNAME (sTOP (FloorListTemp)) = v; 
	      ++numregions;
	    };

	  region[v] = region[snNAME (sTOP (FloorListTemp))];
	  region[snNAME (sTOP (FloorListTemp))] = v;

	}
    }

}


/*******************************Query-cdequiv******************************/

void CdequivQuery(int i, FILE *fpo, BOOLEAN printp)
{ int next = region[i];
  if (next)
    {if (printp) fprintf(fpo, "Node %d has cdeq class = { %d", i, i);
     while (next != i)
       {if (printp) fprintf(fpo, ",%d ", next);
	next = region[next];
      }
     if (printp) fprintf(fpo,"}\n");	 
   }
  else 
    if (printp) fprintf(fpo, "Node %d has EMPTY cdeq class\n", i);
  
}





