/*
 * stack.c - stack abstract data type.  (See stack.h)
 * Copyright (C) 1993 Cornell University -- The Typhoon Project
 */

#include <stdio.h>
#include <assert.h>
#include "cfg.h"
#include "stack.h"

static Stack stack[STMAX];
static Stack *freestack = NULL;

static StackNode stacknode[STMAX];
static StackNode *freestacknode = NULL;

/*
 *   initializes freelists for stacks, stacknodes.
 */
void init_S (void)
{
  Stack *s, *next_s = NULL;
  StackNode *sn, *next_sn = NULL;
  int i;

  for (i=0; i<STMAX; i++)
    {
      s = &stack[i];
      sNEXT (s) = next_s;
      next_s = s;
    }
  freestack = next_s;

  for (i=0; i<STMAX; i++)
    {
      sn = &stacknode[i];
      snNEXT (sn) = next_sn;
      next_sn = sn;
    }
  freestacknode = next_sn;
}

/*
 *   creates a stack.
 */
Stack *create_S ()
{
  Stack *S = freestack;
  freestack = sNEXT (freestack);

  sTOP (S) = NULL;
  sBOT (S) = NULL;
  sSIZE (S) = 0;
  return S;
}

void destroy_S (Stack * S)
{
  sNEXT (S) = freestack;
  freestack = S;
}

/*
 *   adds stacknode to top of stack S; returns added stacknode.
 */
StackNode *push_S (Stack * S, int ei)
{
  StackNode *sn, *oldtop;

  assert (freestacknode != NULL);
  sn = freestacknode;
  freestacknode = snNEXT (freestacknode);

  sSIZE(S) += 1;
  oldtop = sTOP(S);

  snSIZE(sn) = 0;
  snNAME(sn) = 0;
  snNEXT(sn) = NULL;
  snBACK(sn) = oldtop;
  snQNEXT(sn) = NULL;
  snEDGEINDEX(sn) = ei;

  if (oldtop)
    snNEXT(oldtop) = sn;

  sTOP(S) = sn;
  if (sBOT(S) == NULL)
    sBOT(S) = sn;

  return sn;
}

/*
 *   removes stacknode sn from stack S.
 */
void remove_S (StackNode *sn, Stack * S)
{
  sSIZE(S) -= 1;

  if (sn == sTOP(S))
    sTOP(S) = snBACK(sn);
  else
    snBACK(snNEXT(sn)) = snBACK(sn);

  if (sn == sBOT(S))
    sBOT(S) = snNEXT(sn);
  else
    snNEXT(snBACK(sn)) = snNEXT(sn);

  /* add sn to freelist */
  snNEXT (sn) = freestacknode;
  freestacknode = sn;
}

/*
 *   concatenates stack S1 at end of stack S2; destroys stack S1.
 */
void combine_S (Stack * S1, Stack * S2)
{
  assert (!sISEMPTY (S2));

  if (!sISEMPTY (S1))
    {
      sSIZE(S2) += sSIZE(S1);

      snNEXT(sTOP(S2)) = sBOT(S1);
      snBACK(sBOT(S1)) = sTOP(S2);
      sTOP(S2) = sTOP(S1);
    }

  /* add S1 to freelist */
  sNEXT (S1) = freestack;
  freestack = S1;
}

/* end of stack.c */
